/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
 
package ca.uvic.csr.shrimp.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;

import ca.uvic.csr.shrimp.DataBean.DataBean;
import ca.uvic.csr.shrimp.DisplayBean.DisplayBean;
 
/**
 * This class is a dialog that display a request for the user
 * to choose a ChildParent relationship.
 *
 * @author Casey Best
 * date: Nov 6, 2000
 */
public class ChildParentChooserWindow extends JDialog {
    
	private List allCprelTypes = new ArrayList();
    private JCheckBox chkFlattenHierarchy;
	private JButton btnCancel;
	private JButton btnOK;
	private JButton btnRestoreDefault;
	private JList lstArcTypes;
	private boolean okPressed;
	private JCheckBox chkInvert;
	//private boolean inverted = false;

	public ChildParentChooserWindow (Frame frame, DisplayBean displayBean, final DataBean dataBean) {
		super (frame, "Hierarchy Chooser", true);
		
		//this.selectedCprels = displayBean.getCprels();
		okPressed = false;
		
		// add the ok button
		JPanel buttonPanel = new JPanel();

		// the set defaults button
		btnRestoreDefault = new JButton ("Restore Default");
		btnRestoreDefault.setBounds (0, 0, 20, 21);
		btnRestoreDefault.addActionListener ( new ActionListener () {
			public void actionPerformed (ActionEvent event) {
			    setListSelection(Arrays.asList(dataBean.getDefaultCprels()));
			    chkFlattenHierarchy.setSelected(false);
			    chkInvert.setSelected(dataBean.getDefaultCprelsInverted());
                chkInvert.setEnabled(true);
                lstArcTypes.setEnabled(true);
			}
		});
		buttonPanel.add (btnRestoreDefault);
		
		// the ok button
		btnOK = new JButton("OK");
		//btnOK.setBounds (0, 0, 20, 21);
		btnOK.addActionListener ( new ActionListener () {
			public void actionPerformed (ActionEvent event) {
				okPressed = true;
				dispose();
			}
		});
		buttonPanel.add(btnOK);

		// the cancel button
		btnCancel = new JButton("Cancel");
		//btnCancel.setBounds (0, 0, 20, 21);
		btnCancel.addActionListener ( new ActionListener () {
			public void actionPerformed (ActionEvent event) {
				okPressed = false;
				dispose();
			}
		});
		buttonPanel.add(btnCancel);

		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
		
		// Create the list of cprels
		allCprelTypes = dataBean.getHierarchicalRelationshipTypes(false, true);

		String [] currentCprels = displayBean.getCprels();
		lstArcTypes = new JList(allCprelTypes.toArray());
		lstArcTypes.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		List currentCprelsList = Arrays.asList(currentCprels);
		setListSelection(currentCprelsList);
		JScrollPane scrollPane = new JScrollPane(lstArcTypes);
		scrollPane.setPreferredSize(new Dimension (300, 200));
        int [] selectedIndices = lstArcTypes.getSelectedIndices();
        if (selectedIndices.length > 0) {
    		lstArcTypes.ensureIndexIsVisible(selectedIndices[0]);
        }        

		// add the combo box to a pane
        JPanel pnlList = new JPanel(new BorderLayout());
        pnlList.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(Color.GRAY, 1), BorderFactory.createEmptyBorder(10,10,10,10)));
        JLabel label = new JLabel("Please choose arc types to structure hierarchy:");
        pnlList.add(label, BorderLayout.CENTER);
        //pnlList.add(Box.createRigidArea(new Dimension(0,5)), BorderLayout.CENTER);
        pnlList.add(scrollPane, BorderLayout.SOUTH);

		getContentPane().add (pnlList, BorderLayout.CENTER);
		
		chkInvert = new JCheckBox ("Invert Hierarchy");
		chkInvert.setSelected(displayBean.isInverted());
		pnlList.add(chkInvert, BorderLayout.NORTH);
				
		chkFlattenHierarchy = new JCheckBox ("No Hierarchy");
		chkFlattenHierarchy.setFont(new Font(chkFlattenHierarchy.getFont().getName(), Font.BOLD, chkFlattenHierarchy.getFont().getSize()));
		chkFlattenHierarchy.setHorizontalAlignment(SwingConstants.CENTER);
		chkFlattenHierarchy.setBorder(BorderFactory.createEmptyBorder(10,5,10,5));
		chkFlattenHierarchy.addActionListener(new ActionListener () {
            public void actionPerformed(ActionEvent e) {
                chkInvert.setEnabled(!chkFlattenHierarchy.isSelected());
                lstArcTypes.setEnabled(!chkFlattenHierarchy.isSelected());
            }
        });
		chkFlattenHierarchy.setSelected(currentCprels.length == 0);
        chkInvert.setEnabled(!chkFlattenHierarchy.isSelected());
        lstArcTypes.setEnabled(!chkFlattenHierarchy.isSelected());
		getContentPane().add(chkFlattenHierarchy, BorderLayout.NORTH);
		setBackground (Color.white);
		pack();

		// relocate and setVisible
		Toolkit toolkit = Toolkit.getDefaultToolkit();
		Dimension screen = toolkit.getScreenSize();
		int x = (int)(screen.getWidth() - getWidth()) / 2;
		int y = (int)(screen.getHeight() - getHeight()) / 2;
		setLocation (x,y);

		getRootPane().setDefaultButton(btnOK);
		setVisible (true);
	}
	
	/**
     * @param relTypes
     */
    private void setListSelection(List relTypes) {
        int [] indices = new int [relTypes.size()];
		for (int i = 0; i < relTypes.size(); i++) {
			String relType = (String)relTypes.get(i);
			indices[i] = allCprelTypes.indexOf(relType);
		}
		lstArcTypes.setSelectedIndices(indices);
    }

    /**
	 * Returns whether or not the accept button was pressed
	 */
	public boolean accepted () {
		return okPressed;
	}
	
	/**
	 * Returns the chosen Child-Parent Relationships
	 */
	public String [] getCprels() {
	    if (chkFlattenHierarchy.isSelected()) {
	        return new String [0];
	    }
	    Object [] selectedValues = lstArcTypes.getSelectedValues();
	    String [] selectedCprels = new String [selectedValues.length];
	    System.arraycopy(selectedValues, 0, selectedCprels, 0, selectedValues.length);
		return selectedCprels;
	}
	
	public boolean isInverted () {
	    return chkInvert.isSelected();
	}
	
} 
