/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.gui.quickview;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.HeadlessException;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.ButtonGroup;
import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextField;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.border.Border;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import ca.uvic.csr.shrimp.LayoutConstants;
import ca.uvic.csr.shrimp.ShrimpConstants;
import ca.uvic.csr.shrimp.DataBean.DataBean;
import ca.uvic.csr.shrimp.DataDisplayBridge.CompositeArcsManager;
import ca.uvic.csr.shrimp.DataDisplayBridge.DataDisplayBridge;
import ca.uvic.csr.shrimp.DataDisplayBridge.RelTypeGroup;
import ca.uvic.csr.shrimp.DisplayBean.DisplayBean;
import ca.uvic.csr.shrimp.DisplayBean.DisplayConstants;
import ca.uvic.csr.shrimp.DisplayBean.layout.Layout;
import ca.uvic.csr.shrimp.ShrimpApplication.ApplicationAccessor;
import ca.uvic.csr.shrimp.ShrimpApplication.BeanNotFoundException;
import ca.uvic.csr.shrimp.ShrimpApplication.ShrimpToolNotFoundException;
import ca.uvic.csr.shrimp.ShrimpProject.ShrimpProject;
import ca.uvic.csr.shrimp.ShrimpTool.ShrimpTool;
import ca.uvic.csr.shrimp.gui.ChooseRelationshipType;
import ca.uvic.csr.shrimp.gui.ShrimpView.ShrimpView;
import ca.uvic.csr.shrimp.resource.IconFilename;
import ca.uvic.csr.shrimp.resource.ResourceHandler;
import ca.uvic.csr.shrimp.util.BrowserLauncher;
import ca.uvic.csr.shrimp.util.CollapsiblePanel;
import ca.uvic.csr.shrimp.util.GradientPanel;
import ca.uvic.csr.shrimp.util.ShrimpUtils;

/**
 * Displays a GUI which lets the user configure the quick views for the current project.
 * The user can create a new quick view (flat or nested), and can delete existing quick views.
 * Existing views can also be configured differently - for example to change the nesting hierarchy
 * or change the icon.
 *
 * @author Chris Callendar
 */
public class QuickViews implements ListSelectionListener {

	private static final Dimension LABEL_SIZE = new Dimension(80, 24);
	private static final Dimension COMBO_SIZE = new Dimension(140, 24);
	private static final Color BORDER_COLOR = new Color(127, 157, 185);
	private static final Border DETAILS_BORDER = BorderFactory.createEmptyBorder(4, 0, 0, 0);
	private static final Border BORDER = BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(BORDER_COLOR),
																			BorderFactory.createEmptyBorder(5, 5, 5, 5));

	private static final String DEFAULT_LAYOUT = ExpressViewConfigurator.DEFAULT_LAYOUT_MODE;
	// @tag shrimp(fitToNodeLabelling)
	private static final String[] LABEL_MODES = {
		DisplayConstants.LABEL_MODE_SCALE_BY_NODE_SIZE,
		DisplayConstants.LABEL_MODE_FIT_TO_NODE,
		DisplayConstants.LABEL_MODE_WRAP_TO_NODE,
		DisplayConstants.LABEL_MODE_FIXED,
		DisplayConstants.LABEL_MODE_SCALE_BY_LEVEL };
	public static final String DEFAULT_LABEL_MODE = ExpressViewConfigurator.DEFAULT_LABEL_MODE;

	private boolean okPressed;
	private QuickViewManager manager;
	private QuickViewAction selectedAction = null;
	private QuickViewActionListener listener = null;
	private CompositeArcsManager compositeManager = null;

	private JPanel gui;
	private CollapsiblePanel paneDetails;
	private CollapsiblePanel paneTypes;
	private CollapsiblePanel paneNesting;
	private CollapsiblePanel paneComposites;

	private JPanel pnlNodes;
	private JPanel pnlArcs;
	private JPanel pnlHierarchy;
	private JPanel pnlComposites;
	private JPanel pnlDetails;
	private JSplitPane splitNodesArcs;
	private JSplitPane splitCompositeArcs;
	private JSplitPane splMain;
	private JList lstNodes;
	private JList lstArcs;
	private JList lstHierarchy;
	private JList lstQuickViews;
	private JList lstCompositeNodes;
	private JList lstDefaultGroups;
	private JList lstCustomGroups;
	private JList lstDefaultCompositeArcs;
	private JList lstCustomCompositeArcs;
	private DefaultListModel lstModel;
	private JToolBar toolbar;
	private JPanel pnlSave;
	private JPanel pnlTopDetails;
	private JTextField txtName;
	private JComboBox cmbLayout;
	private JComboBox cmbLabelMode;
	private JComboBox cmbIcons;
	private JCheckBox chkInverted;
	private JCheckBox chkNoHierarchy;
	private JCheckBox chkPreviewComposites;

	/**
	 * @param layouts the {@link Layout} objects from the {@link DisplayBean}
	 */
	public QuickViews(QuickViewManager manager) throws HeadlessException {
		this.manager = manager;
		this.okPressed = false;
		this.listener = new QuickViewActionListener();

		initialize();
	}

	/**
	 * Returns the composite arcs manager which will be initialized the first time this method is called.
	 */
	public CompositeArcsManager getCompositeArcsManager() {
		ShrimpProject project = manager.getProject();
		if ((compositeManager == null) && (project != null)) {
			try {
				DisplayBean displayBean = (DisplayBean) project.getTool(ShrimpProject.SHRIMP_VIEW).getBean(ShrimpTool.DISPLAY_BEAN);
				DataDisplayBridge dataDisplayBridge = displayBean.getDataDisplayBridge();
				compositeManager = dataDisplayBridge.getCompositeArcsManager();
			} catch (BeanNotFoundException e) {
				e.printStackTrace();
			} catch (ShrimpToolNotFoundException e) {
				e.printStackTrace();
			}
		}
		return compositeManager;
	}

	public void dispose() {
		if (manager != null) {
			manager.save();
			manager = null;
		}
		compositeManager = null;
		selectedAction = null;
		gui.removeAll();
	}

	public void refresh() {
		selectedAction = null;

		// reset the composite arcs manager
		compositeManager = null;
		getCompositeArcsManager();

		initTypeLists();
		initLayouts();
		loadQuickViews();
	}

	public Component getGUI() {
		return gui;
	}

	public void valueChanged(ListSelectionEvent e) {
		if (!e.getValueIsAdjusting()) {
			JList list = getQuickViewsList();
			promptSaveNewQuickView();
			selectedAction = (QuickViewAction) list.getSelectedValue();
			setDetailsForSelectedAction();
		}
	}

	private void setDetailsForSelectedAction() {
		if (selectedAction != null) {
			ExpressViewConfigurator config = selectedAction.getConfigurator();
			String name = selectedAction.getActionName();
			txtName.setText(name);
			validateQuickViewName();
			getDetailsTaskPane().setTitle(name + " Details");
			getIconsComboBox().setSelectedItem(new IconFilename(selectedAction.getIconFilename(), selectedAction.getIcon()));
			getLayoutComboBox().setSelectedItem(config.getLayoutMode());
			getLabelModeComboBox().setSelectedItem(config.getLabelMode());
			setSelectionToList(getNodesList(), config.getNodeTypesOfInterest());
			setSelectionToList(getArcsList(), config.getArcTypesOfInterest());
			setSelectionToList(getHierarchyList(), config.getCprels());
			setSelectionToList(getCompositeNodesList(), config.getNodeTypesToOpen());

			Vector relTypeGroups = new Vector(0);
			CompositeArcsManager mgr = getCompositeArcsManager();
			if (mgr != null) {
				relTypeGroups = mgr.getRelTypeGroups();
				for (Iterator iter = config.getCompositeArcs().keySet().iterator(); iter.hasNext(); ) {
					String groupName = (String) iter.next();
					RelTypeGroup group = new RelTypeGroup(groupName);
					if (!relTypeGroups.contains(group)) {
						relTypeGroups.add(group);
					} else {
						System.out.println("Already contains rel type group: " + groupName);
					}
				}
			}
			getDefaultGroupsList().setListData(relTypeGroups);
			if (relTypeGroups.size() > 0) {
				getDefaultGroupsList().setSelectedIndex(0);
			}
			loadCustomGroups(config);

			if (config.isInverted() != chkInverted.isSelected()) {
				chkInverted.doClick();		// triggers an action event
			}
			if (!selectedAction.isNested() != chkNoHierarchy.isSelected()) {
				chkNoHierarchy.doClick();	// triggers an action event
			}
		} else {
			txtName.setText("");
		}
	}

	private void loadCustomGroups(ExpressViewConfigurator config) {
		Vector customGroups = new Vector(config.getCompositeArcs().keySet());
		Collections.sort(customGroups);
		getCustomGroupsList().setListData(customGroups);
		if (customGroups.size() > 0) {
			getCustomGroupsList().setSelectedIndex(0);
		}
	}

	private void initTypeLists() {
		ShrimpProject project = manager.getProject();
		if (project != null) {
			try {
				DataBean dataBean = (DataBean) project.getBean(ShrimpProject.DATA_BEAN);
				Vector nodes = dataBean.getArtifactTypes(true, true);
				Vector rels = dataBean.getRelationshipTypes(true, true);
				getNodesList().setListData(nodes);
				getArcsList().setListData(rels);
				getHierarchyList().setListData(rels);
				getCompositeNodesList().setListData(nodes);
			} catch (BeanNotFoundException e) {
				e.printStackTrace();
			}
		}
	}

	private void loadQuickViews() {
		// reset the selected action
		selectedAction = null;

		// reset the model
		lstModel.clear();

		QuickViewAction[] quickViews = manager.getSortedQuickViews();
		if (quickViews.length == 0) {
			System.out.println("Loading default quick views");
			manager.loadDefaults();
			quickViews = manager.getSortedQuickViews();
		}
		for (int i = 0; i < quickViews.length; i++) {
			Action action = quickViews[i];
			lstModel.addElement(action);
		}
		if (lstModel.size() > 0) {
			getQuickViewsList().setSelectedIndex(0);
		} else {
			setDetailsForSelectedAction();
		}
	}

	/**
	 * Loads all the icons from the <code>resource/icons</code> directory that start with
	 * <code>icon_quick_view</code>.  If no icons are found then the icons combobox is disabled.
	 */
	private void loadIcons() {
		try {
			IconFilename[] icons = ResourceHandler.getIcons("icon_quick_view");
			getIconsComboBox().setModel(new DefaultComboBoxModel(icons));
		} catch (Exception e) {
			System.err.println("Warning - couldn't load icons: " + e.getMessage());
		}

		// @tag Shrimp(QuickViews(Icons)) : if no icons were loaded - disable the icon combobox?
		if (getIconsComboBox().getModel().getSize() == 0) {
			getIconsComboBox().setEnabled(false);
		}
		/*
		if (getIconsComboBox().getModel().getSize() == 0) {
			// load a default icon
			try {
				BufferedImage img = new BufferedImage(50, 52, BufferedImage.TYPE_BYTE_BINARY);
				Graphics g = img.createGraphics();
				g.setColor(Color.white);
				g.fillRect(0, 0, 50, 52);
				ImageIcon icon = new ImageIcon(img);
				g.dispose();
				IconFilename iconFilename = new IconFilename("icon_quick_view_blank.gif", icon);
				getIconsComboBox().setModel(new DefaultComboBoxModel(new IconFilename[] { iconFilename }));
			} catch (Exception e) {
				System.err.println("Couldn't create a default blank icon: " + e.getMessage());
			}
		}
		*/
 	}

	/**
	 * Sort the layout names and add them to the layout combobox, then select the default layout
	 * or the first layout.
	 */
	private void initLayouts() {
		Vector layoutNames = new Vector();
		// load the layouts from the display bean
		ShrimpProject project = manager.getProject();
		if (project != null) {
			try {
				ShrimpView shrimpView = (ShrimpView) project.getTool(ShrimpProject.SHRIMP_VIEW);
				DisplayBean displayBean = (DisplayBean) shrimpView.getBean(ShrimpTool.DISPLAY_BEAN);
				Vector layouts = displayBean.getLayouts();
				layoutNames = new Vector(layouts.size());
				for (Iterator iter = layouts.iterator(); iter.hasNext(); ) {
					Layout layout = (Layout) iter.next();
					layoutNames.add(layout.getName());
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			// add some default layouts
			layoutNames.add(LayoutConstants.LAYOUT_GRID_BY_TYPE);
			layoutNames.add(LayoutConstants.LAYOUT_RADIAL);
			layoutNames.add(LayoutConstants.LAYOUT_SPRING);
			layoutNames.add(LayoutConstants.LAYOUT_TREE_VERTICAL);
			layoutNames.add(LayoutConstants.LAYOUT_TREE_HORIZONTAL);
		}

		// don't allow users to choose the motion layout
		layoutNames.remove(LayoutConstants.LAYOUT_MOTION);

		if (layoutNames.size() > 0) {
			// sort names alphabetically
			Collections.sort(layoutNames, String.CASE_INSENSITIVE_ORDER);
			// add them to the combobox
			getLayoutComboBox().setModel(new DefaultComboBoxModel(layoutNames));

			if (layoutNames.contains(DEFAULT_LAYOUT)) {
				getLayoutComboBox().setSelectedItem(DEFAULT_LAYOUT);
			} else  {
				getLayoutComboBox().setSelectedIndex(0);
			}
		}
	}

	/**
	 * If the save button is visible then the user is prompted to save the new quick view.
	 */
	private void promptSaveNewQuickView() {
		if (pnlSave.isVisible()) {
			int choice = JOptionPane.showConfirmDialog(gui, "Do you want to save the new quick view?", "Save?", JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
			if ((choice == JOptionPane.YES_OPTION) && !saveNewQuickView()) {
				// revert back to the last selection
				JList list = getQuickViewsList();
				list.removeListSelectionListener(this);
				list.setSelectedValue(selectedAction, false);
				list.addListSelectionListener(this);
			} else if (choice == JOptionPane.NO_OPTION) {
				pnlSave.setVisible(false);
				// remove the temporary new action
				lstModel.removeElement(selectedAction);
				selectedAction = null;
			} else if (choice == JOptionPane.CANCEL_OPTION) {
				// do nothing
			}
		}
	}

	/**
	 * Saves the new quick view.  If the name is blank then an error dialog is displayed and the
	 * quick view is not saved.  The save button is hidden if the quick view is successfully saved.
	 * The new quick view is also selected in the list.
	 * @return true if the quick view was saved.
	 */
	private boolean saveNewQuickView() {
		if (selectedAction == null) {
			return true;
		}

		String name = selectedAction.getActionName();
		ExpressViewConfigurator config = selectedAction.getConfigurator();
		if (name.length() == 0) {
			String msg = "Please enter a name for this quick view.";
			JOptionPane.showMessageDialog(gui, msg, "Error", JOptionPane.ERROR_MESSAGE);
			txtName.requestFocus();
			txtName.setBackground(Color.yellow);
			return false;
		}

		// check that at least one node of interest is selected
		Collection nodes = config.getNodeTypesOfInterest();
		if (nodes.size() == 0) {
			String msg = "This quick view has no nodes of interest and therefore will not be displayed in the quick view panel.\nDo you want to continue?";
			int choice = JOptionPane.showConfirmDialog(gui, msg, "Confirm", JOptionPane.YES_NO_OPTION);
			if (choice == JOptionPane.NO_OPTION) {
				getNodesList().requestFocus();
				getNodesList().setBackground(Color.yellow);
				return false;
			}
		}

		Collection arcs = config.getArcTypesOfInterest();
		String[] cprels = config.getCprels();
		boolean inverted = chkInverted.isSelected();
		String layout = config.getLayoutMode();
		String label = config.getLabelMode();
		String filename = selectedAction.getIconFilename();
		Collection nodesToOpen = config.getNodeTypesToOpen();
		Map cArcs = config.getCompositeArcs();

		QuickViewAction action = QuickViewAction.createGenericView(name, manager.getProject(), nodes, arcs, layout, label, cprels, inverted, nodesToOpen, cArcs);
		action.setIconFilename(filename);
		pnlSave.setVisible(false);
		manager.addQuickView(action);
		loadQuickViews();
		getQuickViewsList().setSelectedValue(action, true);
		return true;
	}

	private void newQuickView() {
		NewQuickViewDialog dlg = new NewQuickViewDialog(getToolBar());
		if (dlg.okPressed) {
			// create and add a dummy action
			QuickViewAction action;
			if (dlg.isNested()) {
				action = QuickViewAction.createNestedView(dlg.getQuickViewName(), manager.getProject());
				action.setIconFilename("icon_quick_view_nested.gif");
			} else {
				action = QuickViewAction.createFlatView(dlg.getQuickViewName(), manager.getProject());
				action.setIconFilename("icon_quick_view_call_graph.gif");
			}
			lstModel.addElement(action);
			getQuickViewsList().setSelectedValue(action, true);
			txtName.requestFocus();
			txtName.selectAll();
			pnlSave.setVisible(true);
		}
	}

	private void deleteSelectedQuickView() {
		// prompt user?

		// if this is a new quick view then don't prompt to save it
		if (pnlSave.isVisible()) {
			pnlSave.setVisible(false);
		}
		selectedAction = null;
		QuickViewAction action = (QuickViewAction) getQuickViewsList().getSelectedValue();
		manager.removeQuickView(action);
		loadQuickViews();
	}

	/**
	 * Checks if a new quick view needs to be saved.
	 */
	public void close() {
		promptSaveNewQuickView();
		manager.save();
	}

	private void initialize() {
		gui = new JPanel(new BorderLayout());
		gui.add(getToolBar(), BorderLayout.NORTH);
		gui.add(getMainSplitPane(), BorderLayout.CENTER);

		loadIcons();
		refresh();

		getMainSplitPane().setDividerLocation(220);
	}

	private JToolBar getToolBar() {
		if (toolbar == null) {
			toolbar = new JToolBar();
			toolbar.setFloatable(false);
			toolbar.setBorder(BorderFactory.createEtchedBorder(1));
			toolbar.add(createNewQuickViewButton());
			toolbar.add(createDeleteQuickViewButton());
			toolbar.add(createRunQuickViewButton());
			toolbar.add(Box.createHorizontalGlue());
			toolbar.add(createHelpButton());
		}
		return toolbar;
	}

	private JButton createHelpButton() {
		JButton btn = new JButton(new AbstractAction(ShrimpConstants.MENU_HELP, ResourceHandler.getIcon("icon_help.gif")) {
			public void actionPerformed(ActionEvent e) {
				try {
					BrowserLauncher.openURL(ShrimpConstants.SHRIMP_QUICKVIEWS_WEBSITE);
				} catch (IOException e1) {
				}
			}
		});
		btn.setToolTipText("Displays extra information about quick views");
		return btn;
	}

	private JButton createRunQuickViewButton() {
		JButton btn = new JButton(new AbstractAction("Run Quick View ", ResourceHandler.getIcon("icon_forward.gif")) {
			public void actionPerformed(ActionEvent e) {
				if (selectedAction != null) {
					selectedAction.startAction();
				}
			}
		});
		btn.setToolTipText("Run the selected quick view in the main view");
		return btn;
	}

	private JButton createDeleteQuickViewButton() {
		JButton btn = new JButton(new AbstractAction("Delete ", ResourceHandler.getIcon("icon_delete.gif")) {
			public void actionPerformed(ActionEvent e) {
				deleteSelectedQuickView();
			}
		});
		btn.setToolTipText("Delete the selected quick view (can't undo!)");
		return btn;
	}

	private JButton createNewQuickViewButton() {
		JButton btn = new JButton(new AbstractAction("New Quick View ", ResourceHandler.getIcon("icon_new.gif")) {
			public void actionPerformed(ActionEvent e) {
				newQuickView();
			}
		});
		btn.setToolTipText("Create a new quick view");
		return btn;
	}

	private JSplitPane getMainSplitPane() {
		if (splMain == null) {
			splMain = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
			splMain.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
			JScrollPane scroll = new JScrollPane(getQuickViewsList());
			scroll.setBorder(BORDER);
			splMain.setLeftComponent(scroll);
			splMain.setRightComponent(getDetailsPanel());
		}
		return splMain;
	}

	private JList getQuickViewsList() {
		if (lstQuickViews == null) {
			lstModel = new DefaultListModel();
			lstQuickViews = new JList(lstModel);
			lstQuickViews.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			lstQuickViews.addListSelectionListener(this);
			lstQuickViews.setCellRenderer(new ActionListCellRenderer());
		}
		return lstQuickViews;
	}

	private JPanel getDetailsPanel() {
		if (pnlDetails == null) {
			pnlDetails = new JPanel(new BorderLayout());
			pnlDetails.setBorder(BorderFactory.createLoweredBevelBorder());

			GradientPanel container = new GradientPanel(new GridBagLayout());
			//container.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
			pnlDetails.add(new JScrollPane(container), BorderLayout.CENTER);
			GridBagConstraints c = new GridBagConstraints();
			c.insets = new Insets(5, 5, 5, 5);
			c.gridx = 0;
			c.gridy = GridBagConstraints.RELATIVE;
			c.fill = GridBagConstraints.HORIZONTAL;
			c.weightx = 1.0;
			container.add(getDetailsTaskPane(), c);
			container.add(getTypesTaskPane(), c);
			container.add(getNestingTaskPane(), c);
			container.add(getCompositesTaskPane(), c);

			c.fill = GridBagConstraints.BOTH;
			c.weighty = 1.0;
			JPanel pnl = new JPanel();
			pnl.setOpaque(false);
			container.add(pnl, c);

			pnlSave = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 1));
			pnlSave.setOpaque(false);
			pnlSave.setPreferredSize(new Dimension(300, 26));
			JButton btnSave = new JButton("Save", ResourceHandler.getIcon("icon_save.gif"));
			btnSave.setPreferredSize(new Dimension(100, 24));
			btnSave.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					saveNewQuickView();
				}
			});
			pnlSave.add(btnSave);
			pnlSave.setVisible(false);
			pnlDetails.add(pnlSave, BorderLayout.SOUTH);
		}
		return pnlDetails;
	}

	private CollapsiblePanel getDetailsTaskPane() {
		if (paneDetails == null) {
			paneDetails = new CollapsiblePanel("Quick View Details", getTopDetailsPanel());
			paneDetails.setCollapsed(false);
		}
		return paneDetails;
	}

	private CollapsiblePanel getTypesTaskPane() {
		if (paneTypes == null) {
			paneTypes = new CollapsiblePanel("Node & Arc Types", getNodesAndArcsPanel());
			paneTypes.setCollapsed(true);
		}
		return paneTypes;
	}

	private CollapsiblePanel getNestingTaskPane() {
		if (paneNesting == null) {
			paneNesting = new CollapsiblePanel("Nesting Hierarcy", getHierarchyPanel());
			paneNesting.setCollapsed(true);
		}
		return paneNesting;
	}

	private CollapsiblePanel getCompositesTaskPane() {
		if (paneComposites == null) {
			paneComposites = new CollapsiblePanel("Composite Arcs", getCompositesPanel());
			paneComposites.setCollapsed(true);
		}
		return paneComposites;
	}

	private JPanel getTopDetailsPanel() {
		if (pnlTopDetails == null) {
			pnlTopDetails = new JPanel(new GridBagLayout());
			pnlTopDetails.setOpaque(false);
			// create a 3x3 grid
			GridBagConstraints c = new GridBagConstraints();
			c.insets = new Insets(2, 2, 2, 4);
			c.fill = GridBagConstraints.HORIZONTAL;
			c.weightx = 0.2;
			c.weighty = 1;
			JLabel lbl = new JLabel("Name: ");
			lbl.setPreferredSize(LABEL_SIZE);
			pnlTopDetails.add(lbl, c);
			c.gridy = 1;
			lbl = new JLabel("Layout: ");
			lbl.setPreferredSize(LABEL_SIZE);
			pnlTopDetails.add(lbl, c);
			c.gridy = 2;
			lbl = new JLabel("Label mode: ");
			lbl.setPreferredSize(LABEL_SIZE);
			pnlTopDetails.add(lbl, c);

			c.gridx = 1;
			c.gridy = 0;
			c.weightx = 0.6;
			txtName = new JTextField();
			txtName.setPreferredSize(COMBO_SIZE);
			txtName.addKeyListener(listener);
			pnlTopDetails.add(txtName, c);
			c.gridy = 1;
			pnlTopDetails.add(getLayoutComboBox(), c);
			c.gridy = 2;
			pnlTopDetails.add(getLabelModeComboBox(), c);

			c.gridx = 2;
			c.gridy = 0;
			c.weightx = 0.2;
			lbl = new JLabel("Icon", JLabel.CENTER);
			pnlTopDetails.add(lbl, c);
			c.gridy = 1;
			c.gridheight = GridBagConstraints.REMAINDER;
			pnlTopDetails.add(getIconsComboBox(), c);
		}
		return pnlTopDetails;
	}

	private JComboBox getIconsComboBox() {
		if (cmbIcons == null) {
			cmbIcons = new JComboBox();
			cmbIcons.setEditable(false);
			cmbIcons.setPreferredSize(new Dimension(73, 56));
			cmbIcons.setMinimumSize(new Dimension(73, 56));
			cmbIcons.setRenderer(new IconListCellRenderer());
			cmbIcons.addActionListener(listener);
		}
		return cmbIcons;
	}

	private JComboBox getLayoutComboBox() {
		if (cmbLayout == null) {
			cmbLayout = new JComboBox();
			cmbLayout.setEditable(false);
			cmbLayout.setPreferredSize(COMBO_SIZE);
			cmbLayout.addActionListener(listener);
		}
		return cmbLayout;
	}

	private JComboBox getLabelModeComboBox() {
		if (cmbLabelMode == null) {
			cmbLabelMode = new JComboBox(LABEL_MODES);
			cmbLabelMode.setSelectedItem(DEFAULT_LABEL_MODE);
			cmbLabelMode.setEditable(false);
			cmbLabelMode.setPreferredSize(COMBO_SIZE);
			cmbLabelMode.addActionListener(listener);
		}
		return cmbLabelMode;
	}

	private JSplitPane getNodesAndArcsPanel() {
		if (splitNodesArcs == null) {
			splitNodesArcs = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
			splitNodesArcs.setOpaque(false);
			splitNodesArcs.setTopComponent(getNodesPanel());
			splitNodesArcs.setBottomComponent(getArcsPanel());
			splitNodesArcs.setDividerLocation(120);
		}
		return splitNodesArcs;
	}

	private JPanel getNodesPanel() {
		if (pnlNodes == null) {
			pnlNodes = new JPanel(new BorderLayout());
			pnlNodes.setOpaque(false);
			pnlNodes.setBorder(DETAILS_BORDER);
			JComponent header = createListHeader("Select the node types of interest: ", getNodesList(), true, true);
			pnlNodes.add(header, BorderLayout.NORTH);
			pnlNodes.add(new JScrollPane(getNodesList()), BorderLayout.CENTER);
		}
		return pnlNodes;
	}

	private JList getNodesList() {
		if (lstNodes == null) {
			lstNodes = new JList(new DefaultListModel());
			lstNodes.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
			lstNodes.addListSelectionListener(listener);
		}
		return lstNodes;
	}

	private JPanel getArcsPanel() {
		if (pnlArcs == null) {
			pnlArcs = new JPanel(new BorderLayout());
			pnlArcs.setOpaque(false);
			pnlArcs.setBorder(DETAILS_BORDER);
			pnlArcs.add(createListHeader("Select the arc types of interest:", getArcsList(), true, true), BorderLayout.NORTH);
			pnlArcs.add(new JScrollPane(getArcsList()), BorderLayout.CENTER);
		}
		return pnlArcs;
	}

	private JList getArcsList() {
		if (lstArcs == null) {
			lstArcs = new JList(new DefaultListModel());
			lstArcs.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
			lstArcs.addListSelectionListener(listener);
		}
		return lstArcs;
	}

	private JPanel getHierarchyPanel() {
		if (pnlHierarchy == null) {
			pnlHierarchy = new JPanel(new BorderLayout());
			pnlHierarchy.setOpaque(false);
			pnlHierarchy.setBorder(DETAILS_BORDER);

			JComponent comp = createListHeader("Select the nesting hierarchy types:");
			comp.add(createButton(createDefaultCprelsAction(), false));
			comp.add(createButton(createSelectNoneAction(getHierarchyList()), false));
			pnlHierarchy.add(comp, BorderLayout.NORTH);
			pnlHierarchy.add(new JScrollPane(getHierarchyList()), BorderLayout.CENTER);

			JPanel pnl = new JPanel(new GridLayout(1, 2));
			pnl.setOpaque(false);
			pnlHierarchy.add(pnl, BorderLayout.SOUTH);
			chkNoHierarchy = new JCheckBox("No hierarchy", false);
			chkNoHierarchy.setOpaque(false);
			pnl.add(chkNoHierarchy);
			chkNoHierarchy.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					boolean noHierarchy = chkNoHierarchy.isSelected();
					getHierarchyList().setEnabled(!noHierarchy);
					//getTabbedPane().setEnabledAt(2, !noHierarchy);
					chkInverted.setEnabled(!noHierarchy);
					if (selectedAction != null) {
						String[] cprels = (noHierarchy ? new String[0] : getSelectionArrayFromList(getHierarchyList()));
						selectedAction.getConfigurator().setCprels(cprels);
					}
				}
			});

			chkInverted = new JCheckBox("Invert hierarchy?");
			chkInverted.setOpaque(false);
			pnl.add(chkInverted);
			chkInverted.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (selectedAction != null) {
						selectedAction.getConfigurator().setInverted(chkInverted.isSelected());
					}
				}
			});
		}
		return pnlHierarchy;
	}

	/** Creates an action for selecting the default cprels. */
	private Action createDefaultCprelsAction() {
		Action defaultAction = new AbstractAction("Default") {
			public void actionPerformed(ActionEvent e) {
				if (selectedAction != null) {
					String[] cprels = selectedAction.getDataBean().getDefaultCprels();
					setSelectionToList(getHierarchyList(), cprels);
					if (((cprels.length > 0) && chkNoHierarchy.isSelected()) ||
						((cprels.length == 0) && !chkNoHierarchy.isSelected())) {
						chkNoHierarchy.doClick();
					}
				}
			}
		};
		defaultAction.putValue(Action.SHORT_DESCRIPTION, "Select the default hierarchy relationships (for nesting)");
		return defaultAction;
	}

	private JList getHierarchyList() {
		if (lstHierarchy == null) {
			lstHierarchy = new JList(new DefaultListModel());
			lstHierarchy.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
			lstHierarchy.addListSelectionListener(listener);
		}
		return lstHierarchy;
	}

	private JPanel getCompositesPanel() {
		if (pnlComposites == null) {
			pnlComposites = new JPanel(new BorderLayout());
			pnlComposites.setOpaque(false);
			JLabel lbl = new JLabel("Composite arcs are only applicable in nested quick views", JLabel.CENTER);
			lbl.setForeground(new Color(196, 0, 0));
			lbl.setBackground(Color.white);
			lbl.setOpaque(true);
			lbl.setBorder(BorderFactory.createLineBorder(Color.lightGray));
			//lbl.setFont(lbl.getFont().deriveFont(Font.ITALIC));
			pnlComposites.add(lbl, BorderLayout.NORTH);
			JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
			splitPane.setOpaque(false);

			JPanel pnl = new JPanel(new BorderLayout());
			pnl.setOpaque(false);
			pnl.setBorder(DETAILS_BORDER);
			pnl.add(createListHeader("Select the nodes types to open:", getCompositeNodesList(), true, true), BorderLayout.NORTH);
			pnl.add(new JScrollPane(getCompositeNodesList()), BorderLayout.CENTER);
			pnl.setPreferredSize(new Dimension(400, 120));
			splitPane.setTopComponent(pnl);

			JSplitPane compositesArcsPanel = getCompositesArcsPanel();
			compositesArcsPanel.setPreferredSize(new Dimension(400, 150));
			splitPane.setBottomComponent(compositesArcsPanel);
			splitPane.setDividerLocation(120);

			pnlComposites.add(splitPane, BorderLayout.CENTER);
		}
		return pnlComposites;
	}

	private JList getCompositeNodesList() {
		if (lstCompositeNodes == null) {
			lstCompositeNodes = new JList(new DefaultListModel());
			lstCompositeNodes.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
			lstCompositeNodes.addListSelectionListener(listener);
		}
		return lstCompositeNodes;
	}

	private JSplitPane getCompositesArcsPanel() {
		if (splitCompositeArcs == null) {
			splitCompositeArcs = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
			splitCompositeArcs.setOpaque(false);

			JComponent header = createListHeader(/* Default */ "Groups:");
			header.add(getPreviewCompositesCheckBox());
			header.setMaximumSize(new Dimension(1000, 18));

			JPanel left = new JPanel(new BorderLayout());
			left.setOpaque(false);
			left.add(header, BorderLayout.NORTH);
			left.add(new JScrollPane(getDefaultGroupsList()), BorderLayout.CENTER);
			splitCompositeArcs.setLeftComponent(left);

			JPanel right = new JPanel(new BorderLayout());
			right.setOpaque(false);
			right.add(createListHeader("Arc types:"), BorderLayout.NORTH);
			right.add(new JScrollPane(getDefaultCompositeArcsList()), BorderLayout.CENTER);
			splitCompositeArcs.setRightComponent(right);

			// Custom Groups - commented out because it is too complicated for the user?
			// the user can create custom groups inside the arc filter panel
			/* Need to change layout into grid?  Used to use GridBag...
			header = createListHeader("Custom Groups:");
			header.add(createButton(createAddCustomGroupAction(), false));
			header.add(createButton(createDeleteCustomGroupAction(), false));
			pnlCompositeArcs.add(header, c);
			header = createListHeader("Arc types:");
			header.add(createButton(createAddCompositeArcTypeAction(), false));
			header.add(createButton(createDeleteCompositeArcTypeAction(), false));
			pnlCompositeArcs.add(header, c);

			pnlCompositeArcs.add(new JScrollPane(getCustomGroupsList()), c);
			pnlCompositeArcs.add(new JScrollPane(getCustomCompositeArcsList()), c);
			*/
			splitCompositeArcs.setDividerLocation(180);
		}
		return splitCompositeArcs;
	}

	private JList getDefaultGroupsList() {
		if (lstDefaultGroups == null) {
			lstDefaultGroups = new JList(new DefaultListModel());
			GroupsListCellRenderer groupsRenderer = new GroupsListCellRenderer(lstDefaultGroups);
			lstDefaultGroups.setCellRenderer(groupsRenderer);
			lstDefaultGroups.registerKeyboardAction(groupsRenderer, KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0), JComponent.WHEN_FOCUSED);
			lstDefaultGroups.addMouseListener(groupsRenderer);
			lstDefaultGroups.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			lstDefaultGroups.addListSelectionListener(listener);
		}
		return lstDefaultGroups;
	}

	private JList getCustomGroupsList() {
		if (lstCustomGroups == null) {
			lstCustomGroups = new JList(new DefaultListModel());
			lstCustomGroups.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			lstCustomGroups.addListSelectionListener(listener);
		}
		return lstCustomGroups;
	}

	private JList getDefaultCompositeArcsList() {
		if (lstDefaultCompositeArcs == null) {
			lstDefaultCompositeArcs = new JList(new DefaultListModel());
			lstDefaultCompositeArcs.setEnabled(false);
		}
		return lstDefaultCompositeArcs;
	}

	private JList getCustomCompositeArcsList() {
		if (lstCustomCompositeArcs == null) {
			lstCustomCompositeArcs = new JList(new DefaultListModel());
			lstCustomCompositeArcs.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
			lstCustomCompositeArcs.addListSelectionListener(listener);
		}
		return lstCustomCompositeArcs;
	}

	private JComponent createListHeader(String title) {
		JToolBar toolbar = new JToolBar();
		toolbar.setFloatable(false);
		toolbar.setOpaque(false);
		toolbar.setBorder(BorderFactory.createEmptyBorder(1,1,1,1));
		toolbar.add(new JLabel(title));
		toolbar.add(Box.createHorizontalGlue());	// right align buttons
		toolbar.setMinimumSize(new Dimension(20, 16));
		toolbar.setPreferredSize(new Dimension(100, 16));
		return toolbar;
	}

	private JComponent createListHeader(String title, JList list, boolean selectAll, boolean selectNone) {
		JComponent header = createListHeader(title);
		if (selectAll) {
			header.add(createButton(createSelectAllAction(list), false));
		}
		if (selectNone) {
			header.add(createButton(createSelectNoneAction(list), false));
		}
		return header;
	}

	private AbstractButton createButton(Action action, boolean toggle) {
		AbstractButton btn;
		if (toggle) {
			btn = new JToggleButton(action);
		} else {
			btn = new JButton(action);
		}
		btn.setOpaque(false);
		btn.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
		String txt = btn.getText();
		String tt = (String)action.getValue(Action.SHORT_DESCRIPTION);
		btn.setToolTipText((tt != null ? tt : txt));
		Dimension dim;
		if (txt == null) {
			// no text - assume a 16x16 icon only
			dim = new Dimension(28, 18);
		} else {
			dim = btn.getPreferredSize();
			dim.setSize(dim.width, 18);
			btn.setForeground(Color.blue);
		}
		btn.setPreferredSize(dim);
		btn.setMaximumSize(new Dimension(dim.width + 12, 18));
		toolbar.add(btn);
		return btn;
	}

	private Action createSelectAllAction(final JList list) {
		Action action = new AbstractAction("Select All") {
			public void actionPerformed(ActionEvent e) {
				if (list.getModel().getSize() > 0) {
					list.setSelectionInterval(0, list.getModel().getSize() - 1);
				}
			}
		};
		action.putValue(Action.SHORT_DESCRIPTION, "Select all the items in the list");
		return action;
	}

	private Action createSelectNoneAction(final JList list) {
		Action action = new AbstractAction("Select None") {
			public void actionPerformed(ActionEvent e) {
				if (list.getModel().getSize() > 0) {
					list.getSelectionModel().clearSelection();
				}
			}
		};
		action.putValue(Action.SHORT_DESCRIPTION, "Select none of the items in the list");
		return action;
	}


	private JCheckBox getPreviewCompositesCheckBox() {
		if (chkPreviewComposites == null) {
			Action previewAction = new AbstractAction("Preview?", null) {
				public void actionPerformed(ActionEvent e) {}
			};
			previewAction.putValue(Action.SHORT_DESCRIPTION, "Preview composite arcs in the main view?");
			chkPreviewComposites = new JCheckBox(previewAction);
			chkPreviewComposites.setOpaque(false);
			chkPreviewComposites.setSelected(true);
			chkPreviewComposites.setPreferredSize(new Dimension(80, 18));
		}
		return chkPreviewComposites;
	}

	protected Action createAddCustomGroupAction() {
		Action action = new AbstractAction("Add") {
			public void actionPerformed(ActionEvent e) {
				if (selectedAction != null) {
					ExpressViewConfigurator config = selectedAction.getConfigurator();
					String newName = JOptionPane.showInputDialog(getCustomGroupsList(), "Enter a name for the new group");
					Map groups = config.getCompositeArcs();
					if (!groups.containsKey(newName)) {
						groups.put(newName, Collections.EMPTY_LIST);
						loadCustomGroups(config);
					} else {
						JOptionPane.showMessageDialog(getCustomGroupsList(), "A group with that name already exists.", "Duplicate group name", JOptionPane.ERROR_MESSAGE);
					}
				}
			}
		};
		action.putValue(Action.SHORT_DESCRIPTION, "Add a new group");
		return action;
	}

	protected Action createDeleteCustomGroupAction() {
		Action action = new AbstractAction(null, ResourceHandler.getIcon("icon_delete.gif")) {
			public void actionPerformed(ActionEvent e) {
				String groupName = (String) getCustomGroupsList().getSelectedValue();
				if ((selectedAction != null) && (groupName != null)) {
					selectedAction.getConfigurator().getCompositeArcs().remove(groupName);
					loadCustomGroups(selectedAction.getConfigurator());
				}
			}
		};
		action.putValue(Action.SHORT_DESCRIPTION, "Delete the selected custom group");
		return action;
	}

	protected Action createAddCompositeArcTypeAction() {
		Action action = new AbstractAction("Add") {
			public void actionPerformed(ActionEvent ae) {
				String groupName = (String) getCustomGroupsList().getSelectedValue();
				if ((manager.getProject() != null) && (selectedAction != null) && (groupName != null)) {
					Map map = selectedAction.getConfigurator().getCompositeArcs();
					Collection cArcs = new HashSet();
					// add existing arc types
					if (map.containsKey(groupName)) {
						cArcs.addAll((Collection)map.get(groupName));
					}

					Frame frame = ApplicationAccessor.getApplication().getParentFrame();
					ChooseRelationshipType dlg = new ChooseRelationshipType(frame, manager.getProject(), true, cArcs);
					if (dlg.okPressed()) {
						cArcs.addAll(dlg.getSelectedRelationships());
						map.put(groupName, cArcs);

						// reset the selection to the group - this will update the composite arcs list
						getCustomGroupsList().setSelectedIndices(new int[0]);
						getCustomGroupsList().setSelectedValue(groupName, false);
					}
				}
			}
		};
		action.putValue(Action.SHORT_DESCRIPTION, "Add a composite arc type");
		return action;
	}

	protected Action createDeleteCompositeArcTypeAction() {
		Action action = new AbstractAction(null, ResourceHandler.getIcon("icon_delete.gif")) {
			public void actionPerformed(ActionEvent e) {
				Object[] cArcs = getCustomCompositeArcsList().getSelectedValues();
				String groupName = (String) getCustomGroupsList().getSelectedValue();
				if ((selectedAction != null) && (cArcs != null) && (groupName != null)) {
					Map map = selectedAction.getConfigurator().getCompositeArcs();
					Collection existingArcs = (Collection) map.get(groupName);
					for (int i = 0; i < cArcs.length; i++) {
						existingArcs.remove(cArcs[i]);
					}
					// reset the selection to the group - this will update the composite arcs list
					getCustomGroupsList().setSelectedIndices(new int[0]);
					getCustomGroupsList().setSelectedValue(groupName, false);
				}
			}
		};
		action.putValue(Action.SHORT_DESCRIPTION, "Delete the selected composite arc type(s)");
		return action;
	}

	/**
	 * Gets a {@link List} of the selected objects in the {@link JList}.
	 * @param lst the list with selected items
	 * @return {@link List} of objects
	 */
	private List getSelectionFromList(JList lst) {
		Object[] sel = lst.getSelectedValues();
		ArrayList list = new ArrayList(sel.length);
		for (int i = 0; i < sel.length; i++) {
			list.add(sel[i]);
		}
		return list;
	}

	/**
	 * Gets an array of the selected objects from the {@link JList}.
	 * @param lst the list with selected items
	 * @return String[]
	 */
	private String[] getSelectionArrayFromList(JList lst) {
		List list = getSelectionFromList(lst);
		return (String[]) list.toArray(new String[list.size()]);
	}

	/**
	 * Selects the String types in the given {@link JList}.
	 * @param lst the list to set the selection on
	 * @param types the values in the list to select
	 */
	private void setSelectionToList(JList lst, String[] types) {
		ArrayList list = new ArrayList(types.length);
		for (int i = 0; i < types.length; i++) {
			list.add(types[i]);
		}
		setSelectionToList(lst, list);
	}

	/**
	 * Utility method for setting the selection to a JList.
	 * @param lst the list to set the selection on
	 * @param types the objects in the list that should be selected
	 */
	private void setSelectionToList(JList lst, Collection types) {
		if (types == null) {
			return;
		}

		int[] indices = new int[0];
		List indicesList = new ArrayList();
		lst.setSelectedIndices(indices); // clear selected
		for (int i = 0; i < lst.getModel().getSize(); i++) {
			Object artType = lst.getModel().getElementAt(i);
			if (types.contains(artType)) {
				indicesList.add(new Integer(i));
			}
		}
		indices = new int[indicesList.size()];
		int i = 0;
		for (Iterator iter = indicesList.iterator(); iter.hasNext();) {
			Integer indexInt = (Integer) iter.next();
			indices[i] = indexInt.intValue();
			i++;
		}
		lst.setSelectedIndices(indices);
		// ensure the first index is visible
		if (indices.length > 0) {
			lst.ensureIndexIsVisible(indices[0]);
		}
	}

	public boolean okPressed() {
		return okPressed;
	}

	private void validateQuickViewName() {
		String txt = txtName.getText().trim();
		if ((txt.length() > 0) && Color.yellow.equals(txtName.getBackground())) {
			txtName.setBackground(Color.white);
		} else if (txt.length() == 0) {
			txtName.setBackground(Color.yellow);
		}

		// update the selctedAction's name
		if (selectedAction != null) {
			if ((txt.length() > 0) && !txt.equals(selectedAction.getActionName())) {
				selectedAction.putValue(Action.NAME, txt);
			}
		}
	}

	public static void showQuickViewsDialog(Frame parent, final QuickViewManager manager) {
		final QuickViews qv = new QuickViews(manager);
		final JDialog dlg = new JDialog(parent, true);
		dlg.getContentPane().add(qv.getGUI(), BorderLayout.CENTER);
		JPanel pnlButtons = new JPanel(new FlowLayout(FlowLayout.CENTER));
		dlg.add(pnlButtons, BorderLayout.SOUTH);

		JButton btnOK = new JButton("OK");
		btnOK.setPreferredSize(new Dimension(84, 24));
		pnlButtons.add(btnOK);
		btnOK.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				qv.close();
				manager.saveBounds(dlg);
				dlg.setVisible(false);
				dlg.dispose();
			}
		});

		JButton btnDefault = new JButton("Defaults");
		btnDefault.setToolTipText("Revert back to the default quick views");
		btnDefault.setPreferredSize(new Dimension(84, 24));
		pnlButtons.add(btnDefault);
		btnDefault.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String msg = "Are you sure you want to revert to the default quick views?\nYou will lose any new quick views that you've created.";
				int choice = JOptionPane.showConfirmDialog(dlg, msg, "Confirm", JOptionPane.YES_NO_OPTION);
				if (choice == JOptionPane.YES_OPTION) {
					qv.manager.loadDefaults();
					qv.loadQuickViews();
				}
			}
		});

		dlg.setTitle("Customize Quick Views");
		dlg.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
		final Dimension dim = new Dimension(700, 650);
		dlg.setPreferredSize(dim);
		dlg.setSize(dim);
		dlg.pack();

		// load and save the bounds
		if (manager != null) {
			manager.loadBounds(dlg);
		}
		dlg.getRootPane().setDefaultButton(btnOK);
		dlg.setVisible(true);
	}

	class QuickViewActionListener extends KeyAdapter implements ListSelectionListener, ActionListener {

		public void valueChanged(ListSelectionEvent e) {
			// fired by the lists (nodes, arcs, hierarchy, ...)
			if ((selectedAction == null) || e.getValueIsAdjusting()) {
				return;
			}

			ExpressViewConfigurator config = selectedAction.getConfigurator();
			Object src = e.getSource();
			if (src == getNodesList()) {
				List sel = getSelectionFromList(getNodesList());
				config.setNodeTypesOfInterest(sel);
				getNodesList().setBackground((sel.size() == 0 ? Color.yellow : Color.white));
			} else if (src == getArcsList()) {
				config.setArcTypesOfInterest(getSelectionFromList(getArcsList()));
			} else if (src == getHierarchyList()) {
				config.setCprels(getSelectionArrayFromList(getHierarchyList()));
			} else if (src == getCompositeNodesList()) {
				config.setNodeTypesToOpen(getSelectionFromList(getCompositeNodesList()));
			} else if (src == getDefaultGroupsList()) {
				setGroupRelTypes(config, (RelTypeGroup) getDefaultGroupsList().getSelectedValue());
			} else if (src == getCustomGroupsList()) {
				setCustomGroupRelTypes(config, (String) getCustomGroupsList().getSelectedValue());
			}
		}

		private void setCustomGroupRelTypes(ExpressViewConfigurator config, String groupName) {
			Collection arcs = (Collection) config.getCompositeArcs().get(groupName);
			Vector v = (arcs == null ? new Vector(0) : (arcs instanceof Vector ? (Vector)arcs : new Vector(arcs)));
			Collections.sort(v);
			getCustomCompositeArcsList().setListData(v);
			if (v.size() > 0) {
				getCustomCompositeArcsList().setSelectedIndex(0);
			}
		}

		private void setGroupRelTypes(ExpressViewConfigurator config, RelTypeGroup group) {
			if (group != null) {
				Collection c = group.getRelTypes();
				if (c.size() == 0) {
					// the size will be 0 for any of the custom groups, and will be > 0 for the default group
					c = (Collection) config.getCompositeArcs().get(group.getGroupName());
					if (c == null) {
						c = Collections.EMPTY_LIST;
					}
				}
				Vector types = (c instanceof Vector ? (Vector)c : new Vector(c));
				getDefaultCompositeArcsList().setListData(types);
			}
		}

		public void actionPerformed(ActionEvent e) {
			// fired by the 3 combo boxes
			if (selectedAction != null) {
				ExpressViewConfigurator config = selectedAction.getConfigurator();
				Object src = e.getSource();
				if (src == getIconsComboBox()) {
					IconFilename icon = (IconFilename) getIconsComboBox().getSelectedItem();
					if ((icon != null) && !icon.getFilename().equalsIgnoreCase(selectedAction.getIconFilename())) {
						selectedAction.setIconFilename(icon.getFilename());
						getQuickViewsList().repaint();	// re-render the list
					}
				} else if (src == getLayoutComboBox()) {
					String newLayout = (String) getLayoutComboBox().getSelectedItem();
					if ((newLayout != null) && !newLayout.equals(config.getLayoutMode())) {
						config.setLayoutMode(newLayout);
					}
				} else if (src == getLabelModeComboBox()) {
					String newLabelMode = (String) getLabelModeComboBox().getSelectedItem();
					if ((newLabelMode != null) && !newLabelMode.equals(config.getLabelMode())) {
						config.setLabelMode(newLabelMode);
					}
				}
			}
		}

		public void keyReleased(KeyEvent e) {
			validateQuickViewName();
		}

	}

	class ActionListCellRenderer extends JLabel implements ListCellRenderer {
		private final Border NO_BORDER = BorderFactory.createEmptyBorder(1, 1, 1, 1);
		private final Border INVALID_BORDER = BorderFactory.createLineBorder(Color.red, 1);
		private final Border VALID_FOCUS_BORDER = BorderFactory.createLineBorder(Color.cyan, 1);
		private final Color INVALID = new Color(204, 204, 204);

		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
			QuickViewAction action = (QuickViewAction) value;
			setText(action.getActionName());
			setIcon(action.getIcon());
			boolean valid = action.isValid();
			if (isSelected) {
				setBackground(list.getSelectionBackground());
		  		setForeground(valid ? list.getSelectionForeground() : list.getSelectionForeground().brighter());
			} else {
				setBackground(valid ? list.getBackground() : INVALID);
				setForeground(valid ? list.getForeground() : list.getForeground().darker());
			}
			setBorder(valid ? (cellHasFocus ? VALID_FOCUS_BORDER : NO_BORDER) : INVALID_BORDER);
			setEnabled(list.isEnabled());
			setFont(list.getFont());
			setOpaque(true);
			return this;
		}
	}

	class IconListCellRenderer extends JLabel implements ListCellRenderer {

		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
			if (value instanceof IconFilename) {
				IconFilename iconFilename = (IconFilename) value;
				Icon icon = iconFilename.getIcon();
				if (icon == null) {
					icon = iconFilename.reloadIcon();
				}
				setIcon(icon);
				String filename = iconFilename.getFilename();
				setText((icon == null ? filename : null));
				setToolTipText(filename);
			}
			if (isSelected) {
				setBackground(list.getSelectionBackground());
		  		setForeground(list.getSelectionForeground());
			} else {
				setBackground(list.getBackground());
				setForeground(list.getForeground());
			}
			setEnabled(list.isEnabled());
			setFont(list.getFont());
			setOpaque(true);
			return this;
		}
	}

	class GroupsListCellRenderer extends JCheckBox implements ListCellRenderer, ActionListener, MouseListener {
		private JList list;
		private final int hotspot;

		public GroupsListCellRenderer(JList list) {
			super();
			this.list = list;
			hotspot = this.getPreferredSize().width;
		}

		public void actionPerformed(ActionEvent e) {
			toggleSelection(list.getSelectedIndex());
		}

	    private void toggleSelection(int index){
	        if (index < 0) {
				return;
			}

	        RelTypeGroup group = (RelTypeGroup) list.getModel().getElementAt(index);
        	CompositeArcsManager mgr = getCompositeArcsManager();
	        if ((selectedAction != null) && (group != null) && (mgr != null)) {
	        	if (chkPreviewComposites.isSelected()) {
	        		mgr.setCompositesEnabled(group, !group.areCompositesEnabled());
	        	}
	        	ExpressViewConfigurator config = selectedAction.getConfigurator();
				Map cArcs = config.getCompositeArcs();
	        	String groupName = group.getGroupName();
	        	boolean add = false;
				if (cArcs.containsKey(groupName)) {
	        		cArcs.remove(groupName);
	        	} else {
	        		cArcs.put(groupName, group.getRelTypes());
	        		add = true;
	        	}
				// repaint this list - this will check/uncheck the checkbox
		        list.repaint(list.getCellBounds(index, index));

		        // now reload the custom groups list and selec the added group
		        loadCustomGroups(config);
		        if (add) {
		        	getCustomGroupsList().setSelectedValue(groupName, true);
		        }
	        }
	    }

	    /*------------------------------[ MouseListener ]-------------------------------------*/

	    public void mouseEntered(MouseEvent e) {}
	    public void mouseExited(MouseEvent e) {}
	    public void mousePressed(MouseEvent e) {}
	    public void mouseReleased(MouseEvent e) {}

	    public void mouseClicked(MouseEvent me) {
	        int index = list.locationToIndex(me.getPoint());
	        if (index < 0) {
				return;
			}

	        // return if the click is not on the checkbox and is a single click
	        // toggle if it is a double click
			if ((me.getX() > list.getCellBounds(index, index).x + hotspot) && (me.getClickCount() == 1)) {
				return;
			}

			toggleSelection(index);
	    }

		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
			RelTypeGroup group = (RelTypeGroup) value;
			String groupName = group.getGroupName();
			setText(groupName);
			setToolTipText("Check the box to show composite arcs for this group and its arc types");
			if (selectedAction != null) {
				// group.areCompositesEnabled()
				setSelected(selectedAction.getConfigurator().getCompositeArcs().containsKey(groupName));
			}
			if (isSelected) {
				setBackground(list.getSelectionBackground());
		  		setForeground(list.getSelectionForeground());
			} else {
				setBackground(list.getBackground());
				setForeground(list.getForeground());
			}
			setEnabled(list.isEnabled());
			setFont(list.getFont());
			setOpaque(true);
			return this;
		}

	}

	class NewQuickViewDialog extends JDialog implements ActionListener {

		private JTextField txtName;
		private JRadioButton btnNested;
		private JRadioButton btnFlat;
		private boolean okPressed;

		public NewQuickViewDialog(Component parent) {
			super(ShrimpUtils.getParentDialog(parent));
			this.okPressed = false;
			setModal(true);
			setTitle("New Quick View");
			setDefaultCloseOperation(DISPOSE_ON_CLOSE);
			getContentPane().setLayout(new BorderLayout());

			JPanel pnl = new JPanel(new FlowLayout(FlowLayout.LEFT));
			pnl.add(new JLabel("Name: "));
			txtName = new JTextField("New Quick View");
			txtName.setPreferredSize(new Dimension(90, 20));
			pnl.add(txtName);
			getContentPane().add(pnl, BorderLayout.NORTH);

			pnl = new JPanel(new GridLayout(2, 1, 0, 0));
			btnNested = new JRadioButton("Nested Quick View", true);
			pnl.add(btnNested);
			btnFlat = new JRadioButton("Flat Quick View");
			pnl.add(btnFlat);
			pnl.setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
			getContentPane().add(pnl, BorderLayout.CENTER);

			ButtonGroup group = new ButtonGroup();
			group.add(btnNested);
			group.add(btnFlat);

			pnl = new JPanel(new FlowLayout(FlowLayout.CENTER));
			final JButton btnOK = new JButton(" OK ");
			btnOK.setActionCommand("OK");
			btnOK.addActionListener(this);
			pnl.add(btnOK);
			JButton btnCancel = new JButton("Cancel");
			btnCancel.setActionCommand("Cancel");
			btnCancel.addActionListener(this);
			pnl.add(btnCancel);
			getContentPane().add(pnl, BorderLayout.SOUTH);

			txtName.addKeyListener(new KeyAdapter() {
				public void keyReleased(KeyEvent e) {
					btnOK.setEnabled(txtName.getText().trim().length() > 0);
				}
			});

			pack();
			txtName.selectAll();
			txtName.requestFocus();
			Dimension dim = new Dimension(160, 160);
			setPreferredSize(dim);
			setSize(dim);
			pack();
			ShrimpUtils.centerWindowOnParent(this, parent);
			getRootPane().setDefaultButton(btnOK);
			setVisible(true);
		}

		public String getQuickViewName() {
			return txtName.getText().trim();
		}

		public boolean okPressed() {
			return okPressed;
		}

		public boolean isNested() {
			return btnNested.isSelected();
		}

		public boolean isFlat() {
			return btnFlat.isSelected();
		}

		public void actionPerformed(ActionEvent e) {
			String cmd = e.getActionCommand();
			if ("OK".equals(cmd)) {
				okPressed = true;
			}
			setVisible(false);
			dispose();
		}
	}

	public static void main(String[] args) {
		showQuickViewsDialog(null, new QuickViewManager(null));
	}

}
